from localstack.utils.aws import aws_models
mthvW=super
mthvN=None
mthvG=id
class LambdaLayer(aws_models.LambdaFunction):
 def __init__(self,arn):
  mthvW(LambdaLayer,self).__init__(arn)
  self.cwd=mthvN
  self.runtime=""
  self.handler=""
  self.envvars={}
  self.versions={}
class BaseComponent(aws_models.Component):
 def name(self):
  return self.mthvG.split(":")[-1]
class RDSDatabase(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(RDSDatabase,self).__init__(mthvG,env=env)
class RDSCluster(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(RDSCluster,self).__init__(mthvG,env=env)
class AppSyncAPI(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(AppSyncAPI,self).__init__(mthvG,env=env)
class AmplifyApp(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(AmplifyApp,self).__init__(mthvG,env=env)
class ElastiCacheCluster(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(ElastiCacheCluster,self).__init__(mthvG,env=env)
class TransferServer(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(TransferServer,self).__init__(mthvG,env=env)
class CloudFrontDistribution(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(CloudFrontDistribution,self).__init__(mthvG,env=env)
class CodeCommitRepository(BaseComponent):
 def __init__(self,mthvG,env=mthvN):
  mthvW(CodeCommitRepository,self).__init__(mthvG,env=env)
# Created by pyminifier (https://github.com/liftoff/pyminifier)



# version of localstack-ext
VERSION = "0.12.13.26"

# TODO: fix this. Also, not sure which timezone AWS uses - should be UTC, but there
# have been examples of AccessToken validation failure because of local time comparison
TOKEN_EXPIRY_SECONDS = 24 * 60 * 60

# name of Docker registry for Lambda images
DEFAULT_LAMBDA_DOCKER_REGISTRY = "localstack/lambda"

# Github repo with various code artifacts
ARTIFACTS_REPO = "https://github.com/localstack/localstack-artifacts"



import getpass
FeLIj=object
FeLIo=staticmethod
FeLIA=False
FeLIM=Exception
FeLIH=None
FeLIB=input
FeLIu=list
import json
import logging
import sys
from localstack.config import CONFIG_FILE_PATH,load_config_file
from localstack.constants import API_ENDPOINT
from localstack.utils.common import load_file,safe_requests,save_file,to_str
LOG=logging.getLogger(__name__)
class AuthProvider(FeLIj):
 @FeLIo
 def name():
  raise
 def get_or_create_token(self,username,password,headers):
  pass
 def get_user_for_token(self,token):
  pass
 @FeLIo
 def providers():
  return{c.name():c for c in AuthProvider.__subclasses__()}
 @FeLIo
 def get(provider,raise_error=FeLIA):
  provider_class=AuthProvider.providers().get(provider)
  if not provider_class:
   msg='Unable to find auth provider class "%s"'%provider
   LOG.warning(msg)
   if raise_error:
    raise FeLIM(msg)
   return FeLIH
  return provider_class()
class AuthProviderInternal(AuthProvider):
 @FeLIo
 def name():
  return "internal"
 def get_or_create_token(self,username,password,headers):
  data={"username":username,"password":password}
  response=safe_requests.post("%s/user/signin"%API_ENDPOINT,json.dumps(data),headers=headers)
  if response.status_code>=400:
   return
  try:
   result=json.loads(to_str(response.content or "{}"))
   return result["token"]
  except FeLIM:
   pass
 def read_credentials(self,username):
  print("Please provide your login credentials below")
  if not username:
   sys.stdout.write("Username: ")
   sys.stdout.flush()
   username=FeLIB()
  password=getpass.getpass()
  return username,password,{}
 def get_user_for_token(self,token):
  raise FeLIM("Not implemented")
def login(provider,username=FeLIH):
 auth_provider=AuthProvider.get(provider)
 if not auth_provider:
  providers=FeLIu(AuthProvider.providers().keys())
  raise FeLIM('Unknown provider "%s", should be one of %s'%(provider,providers))
 username,password,headers=auth_provider.read_credentials(username)
 print("Verifying credentials ... (this may take a few moments)")
 token=auth_provider.get_or_create_token(username,password,headers)
 if not token:
  raise FeLIM("Unable to verify login credentials - please try again")
 configs=load_config_file()
 configs["login"]={"provider":provider,"username":username,"token":token}
 save_file(CONFIG_FILE_PATH,json.dumps(configs))
def logout():
 configs=json_loads(load_file(CONFIG_FILE_PATH,default="{}"))
 configs["login"]={}
 save_file(CONFIG_FILE_PATH,json.dumps(configs))
def json_loads(s):
 return json.loads(to_str(s))
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
sexAY=False
sexAG=True
sexAy=Exception
import os
import traceback
import localstack
from localstack.constants import LOCALSTACK_ROOT_FOLDER
from localstack.utils.common import(chmod_r,download,get_arch,in_docker,is_alpine,is_command_available,mkdir,new_tmp_file,now,rm_rf,run,unzip)
from localstack_ext.constants import ARTIFACTS_REPO
LOG=logging.getLogger(__name__)
RULE_ENGINE_INSTALL_URL="https://github.com/whummer/serverless-iot-offline"
H2_DOWNLOAD_URL="http://www.h2database.com/h2-2019-10-14.zip"
REDIS_URL_PATTERN="https://github.com/whummer/miniredis/raw/master/build/miniredis.<arch>.bin"
SSL_CERT_URL="%s/raw/master/local-certs/server.key"%ARTIFACTS_REPO
MINIREDIS_BINARY=os.path.join(LOCALSTACK_ROOT_FOLDER,"localstack","infra","redis","miniredis.<arch>.bin")
INFRA_DIR=os.path.join(os.path.dirname(localstack.__file__),"infra")
LOCALSTACK_DIR=os.path.dirname(localstack.__file__)
def install_libs():
 install_iot_rule_engine()
 install_postgres()
 install_timescaledb()
 install_redis()
 install_mqtt()
def install_iot_rule_engine():
 target_dir=LOCALSTACK_DIR
 main_file=os.path.join(target_dir,"node_modules","serverless-iot-offline","query.js")
 if not os.path.exists(main_file):
  LOG.info("Installing IoT rule engine. This may take a while.")
  run("cd %s; npm install %s"%(target_dir,RULE_ENGINE_INSTALL_URL))
 return main_file
def install_postgres():
 if not in_docker():
  return
 check_or_install("postgres","postgresql","RDS")
 check_or_install("pg_config","postgresql-dev","RDS")
 if not is_alpine():
  return
 plpython_lib="/usr/local/lib/postgresql/plpython3.so"
 if os.path.exists(plpython_lib):
  return
 run("apk add postgresql-plpython3")
 ext_source="/usr/share/postgresql/extension"
 ext_target="/usr/local/share/postgresql/extension"
 links={plpython_lib:"/usr/lib/postgresql/plpython3.so","%s/plpython3u.control"%ext_target:"%s/plpython3u.control"%ext_source,"%s/plpython3u--1.0.sql"%ext_target:"%s/plpython3u--1.0.sql"%ext_source,"%s/plpython3u--unpackaged--1.0.sql"%ext_target:"%s/plpython3u--unpackaged--1.0.sql"%ext_source}
 for target,source in links.items():
  if os.path.exists(source)and not os.path.exists(target):
   mkdir(os.path.dirname(target))
   run("ln -s %s %s"%(source,target))
def install_timescaledb():
 if not in_docker():
  return
 if os.path.exists("/usr/lib/postgresql/timescaledb.so"):
  return
 check_or_install("gcc","cmake gcc git musl-dev libc-dev","Timestream")
 ts_dir="/tmp/timescaledb"
 tag="2.0.0-rc4"
 run("cd /tmp; git clone https://github.com/timescale/timescaledb.git")
 run("cd %s; git checkout %s; ./bootstrap -DREGRESS_CHECKS=OFF; cd build; make; make install"%(ts_dir,tag))
 rm_rf("/tmp/timescaledb")
def install_redis():
 check_or_install("redis-server","redis","ElastiCache")
 return "redis-server"
def install_mqtt():
 check_or_install("mosquitto","mosquitto","IoT")
 return "mosquitto"
def install_alpine_package(package,api_name):
 if not is_alpine():
  return
 LOG.info("Downloading dependencies for %s API. This may take a while."%api_name)
 run("apk add %s"%package)
def check_or_install(command,package,api):
 if not is_command_available(command):
  install_alpine_package(package,api)
def setup_ssl_cert():
 from localstack.services import generic_proxy
 target_file=generic_proxy.get_cert_pem_file_path()
 cache_duration_secs=6*60*60
 if os.path.exists(target_file):
  mod_time=os.path.getmtime(target_file)
  if mod_time>(now()-cache_duration_secs):
   return
 download_github_artifact(SSL_CERT_URL,target_file)
def download_github_artifact(url,target_file):
 def do_download(url,print_error=sexAY):
  try:
   download(url,target_file)
   return sexAG
  except sexAy as e:
   if print_error:
    LOG.info("Unable to download local test SSL certificate from %s to %s: %s %s"%(url,target_file,e,traceback.format_exc()))
 result=do_download(url)
 if not result:
  url=url.replace("https://github.com","https://cdn.jsdelivr.net/gh")
  url=url.replace("/raw/master/","@master/")
  do_download(url,sexAG)
def install_h2():
 target_dir=os.path.join(INFRA_DIR,"h2")
 if not os.path.exists(target_dir):
  mkdir(target_dir)
  zip_file=new_tmp_file()
  LOG.info("Downloading dependencies for RDS server. This may take a while.")
  download(H2_DOWNLOAD_URL,zip_file)
  unzip(zip_file,target_dir)
  rm_rf(zip_file)
def install_miniredis():
 arch=get_arch()
 bin_path=MINIREDIS_BINARY.replace("<arch>",arch)
 if not os.path.exists(bin_path):
  redis_folder=os.path.dirname(bin_path)
  mkdir(redis_folder)
  url=REDIS_URL_PATTERN.replace("<arch>",arch)
  LOG.debug("Downloading binary from %s"%url)
  download(url,bin_path)
  chmod_r(bin_path,0o755)
 return bin_path
# Created by pyminifier (https://github.com/liftoff/pyminifier)



#!/usr/bin/env python
MRvFA=None
MRvFI=True
MRvFQ=Exception
MRvFH=str
MRvFf=len
MRvFW=isinstance
MRvFi=dict
MRvFJ=hasattr
MRvFO=int
MRvFL=False
MRvFP=bytes
import json
import logging
import os
import shutil
import socket
import struct
import subprocess
import sys
import tempfile
import threading
import traceback
import uuid
import boto3
import requests
from six.moves.BaseHTTPServer import BaseHTTPRequestHandler,HTTPServer
from six.moves.socketserver import ThreadingMixIn
LOG=logging.getLogger("local_daemon")
DEFAULT_PORT_LOCAL_DAEMON=4600
DEFAULT_PORT_LOCAL_DAEMON_ROOT=4601
DEFAULT_PORT_S3=4566
DEFAULT_PORT_EC2=4566
ENDPOINT_S3="http://localhost:%s"%DEFAULT_PORT_S3
ENDPOINT_EC2="http://localhost:%s"%DEFAULT_PORT_EC2
LOCAL_BIND_ADDRESS_PATTERN="192.168.123.*"
USED_BIND_ADDRESSES=[]
MAC_NETWORK_INTERFACE="en0"
BUCKET_MARKER_LOCAL=(os.environ.get("BUCKET_MARKER_LOCAL")or "").strip()or "__local__"
class FuncThread(threading.Thread):
 def __init__(self,func,params=MRvFA):
  threading.Thread.__init__(self)
  self.daemon=MRvFI
  self.params=params
  self.func=func
 def run(self):
  try:
   self.func(self.params)
  except MRvFQ as e:
   log("Error in thread function: %s %s"%(e,traceback.format_exc()))
class ThreadedHTTPServer(ThreadingMixIn,HTTPServer):
 daemon_threads=MRvFI
class RequestHandler(BaseHTTPRequestHandler):
 def do_POST(self):
  self.read_content()
  try:
   result=self.handle_request()
   self.send_response(200)
  except MRvFQ as e:
   error_string=MRvFH(e)
   result=json.dumps({"error":error_string})
   log("Error handling request: %s - %s"%(self.request_json,e))
   self.send_response(500)
  self.send_header("Content-Length","%s"%MRvFf(result)if result else 0)
  self.end_headers()
  if MRvFf(result or ""):
   self.wfile.write(to_bytes(result))
 def handle_request(self):
  request=self.request_json
  result="{}"
  operation=request.get("op","")
  if operation=="getos":
   result={"result":get_os()}
  elif operation=="shell":
   command=request.get("command")
   result=run_shell_cmd(command)
  elif operation=="s3:download":
   result=s3_download(request)
  elif operation.startswith("root:"):
   result=forward_root_request(request)
  elif operation=="kill":
   log("Terminating local daemon process (port %s)"%DEFAULT_PORT_LOCAL_DAEMON)
   os._exit(0)
  else:
   result={"error":'Unsupported operation "%s"'%operation}
  result=json.dumps(result)if MRvFW(result,MRvFi)else result
  return result
 def read_content(self):
  if MRvFJ(self,"data_bytes"):
   return
  content_length=self.headers.get("Content-Length")
  self.data_bytes=self.rfile.read(MRvFO(content_length))
  self.request_json={}
  try:
   self.request_json=json.loads(self.data_bytes)
  except MRvFQ:
   pass
class RequestHandlerRoot(RequestHandler):
 def handle_request(self):
  request=self.request_json
  result="{}"
  operation=request.get("op")
  if operation=="root:ssh_proxy":
   result=start_ssh_forward_proxy(request)
  elif operation=="kill":
   log("Terminating local daemon process (port %s)"%DEFAULT_PORT_LOCAL_DAEMON_ROOT)
   os._exit(0)
  else:
   result={"error":'Unsupported operation "%s"'%operation}
  result=json.dumps(result)if MRvFW(result,MRvFi)else result
  return result
def s3_download(request):
 bucket=request["bucket"]
 key=request["key"]
 tmp_dir=os.environ.get("TMPDIR")or tempfile.mkdtemp()
 target_file=os.path.join(tmp_dir,request.get("file_name")or "s3file.%s"%MRvFH(uuid.uuid4()))
 if not os.path.exists(target_file)or request.get("overwrite"):
  if bucket==BUCKET_MARKER_LOCAL:
   shutil.copy(key,target_file)
  else:
   s3=boto3.client("s3",endpoint_url=ENDPOINT_S3)
   log("Downloading S3 file s3://%s/%s to %s"%(bucket,key,target_file))
   s3.download_file(bucket,key,target_file)
 return{"local_file":target_file}
def forward_root_request(request):
 url="http://localhost:%s"%DEFAULT_PORT_LOCAL_DAEMON_ROOT
 response=requests.post(url,data=json.dumps(request))
 return json.loads(to_str(response.content))
def start_ssh_forward_proxy(options):
 path=os.path.dirname(__file__)
 if path not in sys.path:
  sys.path.append(path)
 from tcp_proxy import server_loop
 port=options.get("port")or get_free_tcp_port()
 host=LOCAL_BIND_ADDRESS_PATTERN.replace("*",MRvFH(MRvFf(USED_BIND_ADDRESSES)+2))
 create_network_interface_alias(host)
 USED_BIND_ADDRESSES.append(host)
 log("Starting local SSH forward proxy, %s:22 -> localhost:%s"%(host,port))
 options={"bind_port":22,"bind_addr":host,"port":port}
 FuncThread(server_loop,options).start()
 return{"host":host,"forward_port":port}
def create_network_interface_alias(address,interface=MRvFA):
 sudo_cmd="sudo"
 if is_mac_os():
  interface=interface or MAC_NETWORK_INTERFACE
  run_cmd("{sudo_cmd} ifconfig {iface} alias {addr}".format(sudo_cmd=sudo_cmd,addr=address,iface=interface))
  return
 if is_linux():
  interfaces=os.listdir("/sys/class/net/")
  interfaces=[i for i in interfaces if ":" not in i]
  for interface in interfaces:
   try:
    iface_addr=get_ip_address(interface)
    log("Found network interface %s with address %s"%(interface,iface_addr))
    assert iface_addr
    assert interface not in["lo"]and not iface_addr.startswith("127.")
    run_cmd("{sudo_cmd} ifconfig {iface}:0 {addr} netmask 255.255.255.0 up".format(sudo_cmd=sudo_cmd,addr=address,iface=interface))
    return
   except MRvFQ as e:
    log("Unable to create forward proxy on interface %s, address %s: %s"%(interface,address,e))
    pass
 raise MRvFQ("Unable to create network interface")
def run_shell_cmd(command):
 try:
  return{"result":run_cmd(command)}
 except MRvFQ as e:
  error_string=MRvFH(e)
  if MRvFW(e,subprocess.CalledProcessError):
   error_string="%s: %s"%(error_string,e.output)
  return{"error":error_string}
def get_os():
 if is_mac_os():
  return "macos"
 if is_linux():
  return "linux"
 return "windows"
def run_cmd(cmd):
 log("Running command: %s"%cmd)
 return to_str(subprocess.check_output(cmd,stderr=subprocess.STDOUT,shell=MRvFI))
def log(*args):
 print(*args)
 sys.stdout.flush()
def is_mac_os():
 try:
  out=to_str(subprocess.check_output("uname -a",shell=MRvFI))
  return "Darwin" in out
 except subprocess.CalledProcessError:
  return MRvFL
def is_linux():
 try:
  out=to_str(subprocess.check_output("uname -a",shell=MRvFI))
  return "Linux" in out
 except subprocess.CalledProcessError:
  return MRvFL
def get_ip_address(ifname):
 import fcntl
 s=socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
 return socket.inet_ntoa(fcntl.ioctl(s.fileno(),0x8915,struct.pack("256s",to_bytes(ifname[:15])))[20:24])
def get_free_tcp_port():
 tcp=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
 tcp.bind(("",0))
 addr,port=tcp.getsockname()
 tcp.close()
 return port
def to_bytes(obj):
 return obj.encode("utf-8")if MRvFW(obj,MRvFH)else obj
def to_str(obj):
 return obj.decode("utf-8")if MRvFW(obj,MRvFP)else obj
def start_server(port,handler):
 kill_server(port)
 try:
  log("Starting local daemon server on port %s"%port)
  httpd=ThreadedHTTPServer(("0.0.0.0",port),handler)
  httpd.serve_forever()
 except MRvFQ:
  log("Local daemon server already running, or port %s not available"%port)
  pass
def kill_server(port):
 try:
  requests.post("http://localhost:%s"%port,data='{"op":"kill"}')
 except MRvFQ:
  pass
def kill_servers():
 kill_server(DEFAULT_PORT_LOCAL_DAEMON)
 kill_server(DEFAULT_PORT_LOCAL_DAEMON_ROOT)
def main():
 logging.basicConfig()
 daemon_type=sys.argv[1]if MRvFf(sys.argv)>1 else "main"
 os.environ["AWS_ACCESS_KEY_ID"]=os.environ.get("AWS_ACCESS_KEY_ID")or "test"
 os.environ["AWS_SECRET_ACCESS_KEY"]=os.environ.get("AWS_SECRET_ACCESS_KEY")or "test"
 if daemon_type=="main":
  start_server(DEFAULT_PORT_LOCAL_DAEMON,RequestHandler)
 elif daemon_type=="root":
  start_server(DEFAULT_PORT_LOCAL_DAEMON_ROOT,RequestHandlerRoot)
 else:
  log("Unexpected local daemon type: %s"%daemon_type)
def start_in_background():
 from localstack.config import TMP_FOLDER
 from localstack.utils.common import run
 log_file=os.path.join(TMP_FOLDER,"localstack_daemon.log")
 LOG.info("Logging local daemon output to %s"%log_file)
 python_cmd=sys.executable
 cmd="%s %s"%(python_cmd,__file__)
 run(cmd,outfile=log_file,asynchronous=MRvFI)
 LOG.info("Attempting to obtain sudo privileges for local daemon of EC2 API "+"(required to start SSH forward proxy on privileged port 22). "+"You may be asked for your sudo password.")
 run("sudo ls",stdin=MRvFI)
 def start_root_daemon(*args):
  cmd="sudo %s %s root >> %s"%(python_cmd,__file__,log_file)
  run(cmd,outfile=log_file,stdin=MRvFI)
 FuncThread(start_root_daemon).start()
if __name__=="__main__":
 main()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import json
PXpbz=object
PXpba=None
PXpbm=Exception
PXpbe=set
PXpbv=property
PXpbf=classmethod
PXpbN=True
PXpbw=len
PXpbO=getattr
PXpbn=type
PXpbs=isinstance
PXpbF=list
PXpbR=False
import logging
import os
import re
import requests
import yaml
from dulwich import porcelain
from dulwich.client import get_transport_and_path_from_url
from dulwich.repo import Repo
from localstack import config
from localstack.constants import API_ENDPOINT
from localstack.utils.common import(clone,cp_r,disk_usage,download,format_number,is_command_available,load_file,mkdir,new_tmp_file,retry,rm_rf,run,safe_requests,save_file,to_bytes,to_str,unzip)
from localstack.utils.testutil import create_zip_file
from localstack_ext.bootstrap.licensing import get_auth_headers
LOG=logging.getLogger(__name__)
PERSISTED_FOLDERS=["api_states","dynamodb","kinesis"]
class CloudPodManager(PXpbz):
 BACKEND="_none_"
 def __init__(self,pod_name=PXpba,config=PXpba):
  self.pod_name=pod_name
  self._pod_config=config
 def push(self):
  raise PXpbm("Not implemented")
 def pull(self):
  raise PXpbm("Not implemented")
 def get_pod_size(self):
  raise PXpbm("Not implemented")
 def restart_container(self):
  LOG.info("Restarting LocalStack instance with updated persistence state - this may take some time ...")
  data={"action":"restart"}
  url="%s/health"%config.get_edge_url()
  try:
   requests.post(url,data=json.dumps(data))
  except requests.exceptions.ConnectionError:
   pass
  def check_status():
   LOG.info("Waiting for LocalStack instance to be fully initialized ...")
   response=requests.get(url)
   content=json.loads(to_str(response.content))
   statuses=[v for k,v in content["services"].items()]
   assert PXpbe(statuses)==PXpbe(["running"])
  retry(check_status,sleep=3,retries=10)
 @PXpbv
 def pod_config(self):
  return self._pod_config or PodConfigManager.pod_config(self.pod_name)
 @PXpbf
 def get(cls,pod_name):
  pod_config=PodConfigManager.pod_config(pod_name)
  backend=pod_config.get("backend")
  for clazz in cls.__subclasses__():
   if clazz.BACKEND==backend:
    return clazz(pod_name=pod_name,config=pod_config)
  raise PXpbm('Unable to find Cloud Pod manager implementation type "%s"'%backend)
 @PXpbf
 def data_dir(cls):
  if not config.DATA_DIR:
   try:
    details=run("%s inspect %s"%(config.DOCKER_CMD,config.MAIN_CONTAINER_NAME))
    details=json.loads(to_str(details))[0]
    mounts=details.get("Mounts")
    env=details.get("Config",{}).get("Env",[])
    data_dir_env=[e for e in env if e.startswith("DATA_DIR=")][0].partition("=")[2]
    data_dir_host=[m for m in mounts if m["Destination"]==data_dir_env][0]["Source"]
    data_dir_host=re.sub(r"^(/host_mnt)?",r"",data_dir_host)
    config.DATA_DIR=data_dir_host
   except PXpbm:
    LOG.warning('''Unable to determine DATA_DIR from LocalStack Docker container - please make sure $MAIN_CONTAINER_NAME is configured properly''')
  if not config.DATA_DIR:
   raise PXpbm("Working with local cloud pods requires $DATA_DIR configuration")
  return config.DATA_DIR
class CloudPodManagerFilesystem(CloudPodManager):
 BACKEND="file"
 def push(self):
  local_folder=self.target_folder()
  data_dir=self.data_dir()
  print('Pushing state of cloud pod "%s" to local folder: %s'%(self.pod_name,local_folder))
  mkdir(local_folder)
  mkdir(data_dir)
  cp_r(data_dir,local_folder)
  print("Done.")
 def pull(self):
  local_folder=self.target_folder()
  data_dir=self.data_dir()
  if not os.path.exists(local_folder):
   print('WARN: Local path of cloud pod "%s" does not exist: %s'%(self.pod_name,local_folder))
   return
  print('Pulling state of cloud pod "%s" from local folder: %s'%(self.pod_name,local_folder))
  mkdir(data_dir)
  cp_r(local_folder,data_dir)
  self.restart_container()
 def target_folder(self):
  local_folder=re.sub(r"^file://","",self.pod_config.get("url",""))
  return local_folder
 def get_pod_size(self):
  target_folder=self.target_folder()
  return disk_usage(target_folder)
class CloudPodManagerManaged(CloudPodManager):
 BACKEND="managed"
 def push(self):
  presigned_url=self.presigned_url("push")
  data_dir=self.data_dir()
  zip_data_content=create_zip_file(data_dir,get_content=PXpbN)
  print('Pushing state of cloud pod "%s" to backend server (%s KB)'%(self.pod_name,format_number(PXpbw(zip_data_content)/1000.0)))
  res=safe_requests.put(presigned_url,data=zip_data_content)
  if res.status_code>=400:
   raise PXpbm("Unable to push pod state to API (code %s): %s"%(res.status_code,res.content))
  print("Done.")
 def pull(self):
  data_dir=self.data_dir()
  presigned_url=self.presigned_url("pull")
  print('Pulling state of cloud pod "%s" from managed storage'%self.pod_name)
  data_dir=self.data_dir()
  mkdir(data_dir)
  zip_path=new_tmp_file()
  download(presigned_url,zip_path)
  unzip(zip_path,data_dir)
  rm_rf(zip_path)
  self.restart_container()
 def presigned_url(self,mode):
  data={"pod_name":self.pod_name,"mode":mode}
  data=json.dumps(data)
  auth_headers=get_auth_headers()
  response=safe_requests.post("%s/cloudpods/data"%API_ENDPOINT,data,headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise PXpbm("Unable to push cloud pod to API (code %s): %s"%(response.status_code,content))
  content=json.loads(to_str(content))
  return content["presignedURL"]
 def get_pod_size(self):
  data_dir=self.data_dir()
  return disk_usage(data_dir)
class CloudPodManagerGit(CloudPodManager):
 BACKEND="git"
 def push(self):
  repo=self.local_repo()
  client,path=self.client()
  branch=to_bytes(self.pod_config.get("branch"))
  remote_location=self.pod_config.get("url")
  try:
   porcelain.pull(repo,remote_location,refspecs=branch)
  except PXpbm as e:
   LOG.info("Unable to pull repo: %s"%e)
  is_empty_repo=b"HEAD" not in repo or repo.refs.allkeys()==PXpbe([b"HEAD"])
  if is_empty_repo:
   LOG.debug("Initializing empty repository %s"%self.clone_dir)
   init_file=os.path.join(self.clone_dir,".init")
   save_file(init_file,"")
   porcelain.add(repo,init_file)
   porcelain.commit(repo,message="Initial commit")
  if branch not in repo:
   porcelain.branch_create(repo,branch,force=PXpbN)
  self.switch_branch(branch)
  for folder in PERSISTED_FOLDERS:
   LOG.info("Copying persistence folder %s to local git repo %s"%(folder,self.clone_dir))
   src_folder=os.path.join(self.data_dir(),folder)
   tgt_folder=os.path.join(self.clone_dir,folder)
   cp_r(src_folder,tgt_folder)
   files=tgt_folder
   if os.path.isdir(files):
    files=[os.path.join(root,f)for root,_,files in os.walk(tgt_folder)for f in files]
   if files:
    porcelain.add(repo,files)
  porcelain.commit(repo,message="Update state")
  porcelain.push(repo,remote_location,branch)
 def pull(self):
  repo=self.local_repo()
  client,path=self.client()
  remote_refs=client.fetch(path,repo)
  branch=self.pod_config.get("branch")
  remote_ref=b"refs/heads/%s"%to_bytes(branch)
  if remote_ref not in remote_refs:
   raise PXpbm('Unable to find branch "%s" in remote git repo'%branch)
  remote_location=self.pod_config.get("url")
  self.switch_branch(branch)
  branch_ref=b"refs/heads/%s"%to_bytes(branch)
  from dulwich.errors import HangupException
  try:
   porcelain.pull(repo,remote_location,branch_ref)
  except HangupException:
   pass
  for folder in PERSISTED_FOLDERS:
   src_folder=os.path.join(self.clone_dir,folder)
   tgt_folder=os.path.join(self.data_dir(),folder)
   cp_r(src_folder,tgt_folder,rm_dest_on_conflict=PXpbN)
  self.restart_container()
 def client(self):
  client,path=get_transport_and_path_from_url(self.pod_config.get("url"))
  return client,path
 def local_repo(self):
  self.clone_dir=PXpbO(self,"clone_dir",PXpba)
  if not self.clone_dir:
   pod_dir_name=re.sub(r"(\s|/)+","",self.pod_name)
   self.clone_dir=os.path.join(config.TMP_FOLDER,"pods",pod_dir_name,"repo")
   mkdir(self.clone_dir)
   if not os.path.exists(os.path.join(self.clone_dir,".git")):
    porcelain.clone(self.pod_config.get("url"),self.clone_dir)
  return Repo(self.clone_dir)
 def switch_branch(self,branch):
  repo=self.local_repo()
  if is_command_available("git"):
   return run("cd %s; git checkout %s"%(self.clone_dir,to_str(branch)))
  branch_ref=b"refs/heads/%s"%to_bytes(branch)
  if branch_ref not in repo.refs:
   branch_ref=b"refs/remotes/origin/%s"%to_bytes(branch)
  repo.reset_index(repo[branch_ref].tree)
  repo.refs.set_symbolic_ref(b"HEAD",branch_ref)
 def get_pod_size(self):
  self.local_repo()
  return disk_usage(self.clone_dir)
class PodConfigManagerMeta(PXpbn):
 def __getattr__(cls,attr):
  def _call(*args,**kwargs):
   result=PXpba
   for manager in cls.CHAIN:
    try:
     tmp=PXpbO(manager,attr)(*args,**kwargs)
     if tmp:
      if not result:
       result=tmp
      elif PXpbs(tmp,PXpbF)and PXpbs(result,PXpbF):
       result.extend(tmp)
    except PXpbm:
     pass
   if result is not PXpba:
    return result
   raise PXpbm('Unable to run operation "%s" for local or remote configuration'%attr)
  return _call
class PodConfigManager(PXpbz,metaclass=PodConfigManagerMeta):
 CHAIN=[]
 @PXpbf
 def pod_config(cls,pod_name):
  pods=PodConfigManager.list_pods()
  pod_config=[pod for pod in pods if pod["pod_name"]==pod_name]
  if not pod_config:
   raise PXpbm('Unable to find config for pod named "%s"'%pod_name)
  return pod_config[0]
class PodConfigManagerLocal(PXpbz):
 CONFIG_FILE=".localstack.yml"
 def list_pods(self):
  local_pods=self._load_config(safe=PXpbN).get("pods",{})
  local_pods=[{"pod_name":k,"state":"Local Only",**v}for k,v in local_pods.items()]
  existing_names=PXpbe([pod["pod_name"]for pod in local_pods])
  result=[pod for pod in local_pods if pod["pod_name"]not in existing_names]
  return result
 def store_pod_metadata(self,pod_name,metadata):
  pass
 def _load_config(self,safe=PXpbR):
  try:
   return yaml.load(to_str(load_file(self.CONFIG_FILE)))
  except PXpbm:
   if safe:
    return{}
   raise PXpbm('Unable to find and parse config file "%s"'%self.CONFIG_FILE)
class PodConfigManagerRemote(PXpbz):
 def list_pods(self):
  result=[]
  auth_headers=get_auth_headers()
  response=safe_requests.get("%s/cloudpods"%API_ENDPOINT,headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise PXpbm("Unable to fetch list of pods from API (code %s): %s"%(response.status_code,content))
  remote_pods=json.loads(to_str(content)).get("cloudpods",[])
  remote_pods=[{"state":"Shared",**pod}for pod in remote_pods]
  result.extend(remote_pods)
  return result
 def store_pod_metadata(self,pod_name,metadata):
  auth_headers=get_auth_headers()
  metadata["pod_name"]=pod_name
  response=safe_requests.post("%s/cloudpods"%API_ENDPOINT,json.dumps(metadata),headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise PXpbm("Unable to store pod metadata in API (code %s): %s"%(response.status_code,content))
  return json.loads(to_str(content))
PodConfigManager.CHAIN.append(PodConfigManagerLocal())
PodConfigManager.CHAIN.append(PodConfigManagerRemote())
def push_state(pod_name,args):
 backend=CloudPodManager.get(pod_name=pod_name)
 pod_config=clone(backend.pod_config)
 pod_config["size"]=backend.get_pod_size()
 PodConfigManager.store_pod_metadata(pod_name,pod_config)
 backend.push()
def pull_state(pod_name,args):
 if not pod_name:
  raise PXpbm("Need to specify a pod name")
 backend=CloudPodManager.get(pod_name=pod_name)
 backend.pull()
def list_pods(args):
 return PodConfigManager.list_pods()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import base64
DVCeo=None
import hmac
import smtplib
from base64 import b64encode
from smtplib import SMTPAuthenticationError,SMTPException
from localstack.utils.common import to_bytes,to_str
def base64_encode(s):
 result=to_str(b64encode(to_bytes(s))).strip()
 return result
class SMTP(smtplib.SMTP):
 def login(self,user,password):
  def encode_cram_md5(challenge,user,password):
   challenge=base64.decodestring(challenge)
   response=user+" "+hmac.HMAC(password,challenge).hexdigest()
   return base64_encode(response)
  def encode_plain(user,password):
   return base64_encode("\0%s\0%s"%(user,password))
  AUTH_PLAIN="PLAIN"
  AUTH_CRAM_MD5="CRAM-MD5"
  AUTH_LOGIN="LOGIN"
  self.ehlo_or_helo_if_needed()
  if not self.has_extn("auth"):
   raise SMTPException("SMTP AUTH extension not supported by server.")
  authlist=self.esmtp_features["auth"].split()
  preferred_auths=[AUTH_CRAM_MD5,AUTH_PLAIN,AUTH_LOGIN]
  authmethod=DVCeo
  for method in preferred_auths:
   if method in authlist:
    authmethod=method
    break
  if authmethod==AUTH_CRAM_MD5:
   tmp=self.docmd("AUTH",AUTH_CRAM_MD5)
   code=tmp[0]
   resp=tmp[1]
   if code==503:
    return(code,resp)
   tmp=self.docmd(encode_cram_md5(resp,user,password))
   code=tmp[0]
   resp=tmp[1]
  elif authmethod==AUTH_PLAIN:
   tmp=self.docmd("AUTH",AUTH_PLAIN+" "+encode_plain(user,password))
   code=tmp[0]
   resp=tmp[1]
  elif authmethod==AUTH_LOGIN:
   encoded_username=base64_encode(user)
   encoded_password=base64_encode(password)
   tmp=self.check_codes(encoded_username,encoded_password)
   code=tmp[0]
   resp=tmp[1]
  elif authmethod is DVCeo:
   raise SMTPException("No suitable authentication method found.")
  if code not in(235,503):
   raise SMTPAuthenticationError(code,resp)
  return(code,resp)
 def check_codes(self,encoded_username,encoded_password):
  AUTH_LOGIN="LOGIN"
  tmp=self.docmd("AUTH",AUTH_LOGIN)
  code=tmp[0]
  resp=tmp[1]
  if code!=334:
   raise SMTPAuthenticationError(code,resp)
  tmp=self.docmd(encoded_username)
  code=tmp[0]
  resp=tmp[1]
  if code!=334:
   raise SMTPAuthenticationError(code,resp)
  tmp=self.docmd(encoded_password)
  code=tmp[0]
  resp=tmp[1]
  if code!=235:
   raise SMTPAuthenticationError(code,resp)
  return code,resp
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import os
import re
import subprocess

import six
from localstack import config as localstack_config
from localstack import constants as localstack_constants

FALSE_STRINGS = localstack_constants.FALSE_STRINGS

ROOT_FOLDER = os.path.realpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))

PROTECTED_FOLDERS = ("services", "utils")

# api server config
API_URL = localstack_constants.API_ENDPOINT

# database connection settings
DB_NAME = "localstack"
DB_USER = os.environ.get("DB_USER") or "localstack"
DB_PASS = os.environ.get("DB_PASS")

# localhost IP address and hostname
LOCALHOST_IP = "127.0.0.1"
LOCAL_HOSTNAME = "localhost.localstack.cloud"

# bind address of local DNS server
DNS_ADDRESS = os.environ.get("DNS_ADDRESS") or "0.0.0.0"

# IP address that AWS endpoints should resolve to in our local DNS server. By default,
# hostnames resolve to 127.0.0.1, which allows to use the LocalStack APIs transparently
# from the host machine. If your code is running in Docker, this should be configured
# to resolve to the Docker bridge network address, e.g., DNS_RESOLVE_IP=172.17.0.1
DNS_RESOLVE_IP = os.environ.get("DNS_RESOLVE_IP") or LOCALHOST_IP

# fallback DNS server to send upstream requests to
DNS_SERVER = os.environ.get("DNS_SERVER")

# SMTP settings (required, e.g., for Cognito)
SMTP_HOST = os.environ.get("SMTP_HOST", "")
SMTP_USER = os.environ.get("SMTP_USER", "")
SMTP_PASS = os.environ.get("SMTP_PASS", "")
SMTP_EMAIL = os.environ.get("SMTP_EMAIL", "")

# whether to automatically start up utility containers (e.g., Spark/Hadoop for EMR, Presto for Athena)
AUTOSTART_UTIL_CONTAINERS = localstack_config.is_env_true("AUTOSTART_UTIL_CONTAINERS")

# optional flags to pass when starting the bigdata Docker container
BIGDATA_DOCKER_FLAGS = os.environ.get("BIGDATA_DOCKER_FLAGS", "")

# Comma-separated list of regex patterns for DNS names to resolve locally.
# Any DNS name not matched against any of the patterns on this whitelist
# will resolve to the real DNS entry, rather than the local one.
DNS_LOCAL_NAME_PATTERNS = os.environ.get("DNS_LOCAL_NAME_PATTERNS", "")

# whether to use static ports and IDs (e.g., cf-<port>) for CloudFormation distributions
CLOUDFRONT_STATIC_PORTS = localstack_config.is_env_true("CLOUDFRONT_STATIC_PORTS")

# whether to enforce IAM policies when processing requests
ENFORCE_IAM = localstack_config.is_env_true("ENFORCE_IAM")

# folder with persistent API backend states
BACKEND_STATES_DIR = os.path.join(localstack_config.DATA_DIR, "api_states")

# whether to require Pro features and exit with a fault if the API key cannot be activated
REQUIRE_PRO = localstack_config.is_env_true("REQUIRE_PRO")

# endpoint URL for kube cluster (defaults to https://<docker_bridge_ip>:6443)
KUBE_ENDPOINT = os.environ.get("KUBE_ENDPOINT", "")


def use_custom_dns():
    return str(DNS_ADDRESS) not in FALSE_STRINGS


# set USE_SSL to true by default
# TODO fix this!
# localstack_config.USE_SSL = os.environ.get('USE_SSL', '').lower().strip() not in FALSE_STRINGS

# backend service ports
DEFAULT_PORT_LOCAL_DAEMON = 4535
DEFAULT_PORT_LOCAL_DAEMON_ROOT = 4534

# port ranges for service instances (e.g., Postgres DBs, ElastiCache clusters, ...)
SERVICE_INSTANCES_PORTS_START = int(os.environ.get("SERVICE_INSTANCES_PORTS_START") or 4510)
SERVICE_INSTANCES_PORTS_END = int(
    os.environ.get("SERVICE_INSTANCES_PORTS_END") or (SERVICE_INSTANCES_PORTS_START + 30)
)

# port for Azure APIs
PORT_AZURE = 12121

# add default service ports (TODO needed?)
localstack_constants.DEFAULT_SERVICE_PORTS["azure"] = PORT_AZURE

# Docker host name resolvable from containers
DOCKER_HOST_NAME = "host.docker.internal"

# Port where Hive/metastore/Spark are available for EMR/Athena
PORT_HIVE_METASTORE = 9083
PORT_HIVE_SERVER = 10000
PORT_SPARK_MASTER = 7077
PORT_SPARK_UI = 4040

if localstack_config.DOCKER_HOST_FROM_CONTAINER == DOCKER_HOST_NAME:
    # special case when we're running tests outside of Docker
    if not localstack_config.in_docker():
        image_name = localstack_constants.DOCKER_IMAGE_NAME
        cmd = "docker run --rm --entrypoint= -it %s bash -c 'ping -c 1 %s'" % (
            image_name,
            DOCKER_HOST_NAME,
        )
        try:
            out = subprocess.check_output(cmd, shell=True)
            out = out.decode("utf-8") if isinstance(out, six.binary_type) else out
            ip = re.match(r"PING[^\(]+\(([^\)]+)\).*", out, re.MULTILINE | re.DOTALL)
            ip = ip and ip.group(1)
            if ip:
                localstack_config.DOCKER_HOST_FROM_CONTAINER = ip
        except Exception:
            # Swallow this error - Docker daemon potentially not running?
            pass

# update variable names that need to be passed as arguments to Docker
localstack_config.CONFIG_ENV_VARS += [
    "SMTP_HOST",
    "SMTP_USER",
    "SMTP_PASS",
    "SMTP_EMAIL",
    "DNS_SERVER",
    "DNS_ADDRESS",
    "ENFORCE_IAM",
    "DNS_RESOLVE_IP",
    "DNS_LOCAL_NAME_PATTERNS",
    "AUTOSTART_UTIL_CONTAINERS",
    "CLOUDFRONT_STATIC_PORTS",
    "REQUIRE_PRO",
    "AZURE",
    "BIGDATA_DOCKER_FLAGS",
    "LOG_LICENSE_ISSUES",
    "KUBE_ENDPOINT",
]

# re-initialize configs in localstack
localstack_config.populate_configs()



import json
import re

from docopt import docopt
from localstack import config
from localstack.utils.common import format_bytes, truncate

from localstack_ext import constants as ext_constants
from localstack_ext.bootstrap import auth


def cmd_login(argv, args):
    """
    Usage:
      localstack login [options]

    Options:
      --provider=<>           OAuth provider for login (default: github)
      --username=<>           OAuth username for login
    """
    args.update(docopt(cmd_login.__doc__.strip(), argv=argv))

    try:
        provider = args["--provider"] or "internal"
        auth.login(provider, args["--username"])
        print("Successfully logged in.")
    except Exception as e:
        raise Exception("Authentication error: %s" % e)


def cmd_logout(argv, args):
    """
    Usage:
      localstack logout
    """
    args.update(docopt(cmd_logout.__doc__.strip(), argv=argv))
    auth.logout()
    print("Successfully logged out.")


def cmd_ci(argv, args):
    """
    Usage:
      localstack ci <subcommand> [options]

    Commands:
      init                  Initialize CI configuration for a new repository
      repos                 List continuous integration repositories

    Options:
      --provider=<>         CI provider (default: travis)
      --repo=<>             Repository identifier
      --token=<>            API access token for CI provider
    """
    args.update(docopt(cmd_ci.__doc__.strip(), argv=argv))

    if args["<subcommand>"] == "repos":
        provider = args["--provider"] or "travis"
        result = auth.get_ci_repos(provider)
        print(result)
    elif args["<subcommand>"] == "init":
        provider = opt_params(args, ["--provider"]) or "travis"
        repo, token = mand_params(args, ["--repo", "--token"])
        auth.init_ci_repo(repo, token, provider=provider)
        print(repo, token)


def cmd_config(argv, args):
    """
    Usage:
      localstack config [ <key> ]
      localstack config <key> <value>

    Arguments:
      <key>             Key of configuration property to read or write
      <value>           Value of configuration property to write
    """
    args.update(docopt(cmd_config.__doc__.strip(), argv=argv))

    if args["<value>"]:
        auth.set_user_config(args["<key>"], args["<value>"])
        print("Successfully updated configuration value")
    else:
        result = auth.retrieve_user_config()
        if args["<key>"]:
            all_values = result
            result = {}
            for key, value in all_values.items():
                if re.match(args["<key>"], key):
                    result[key] = value
        print(json.dumps(result, indent=4))


def cmd_daemons(argv, args):
    """
    Usage:
      localstack daemons [ <subcommand> ]

    Commands:
      start        Start local daemon processes
      stop         Stop local daemon processes
    """
    args.update(docopt(cmd_daemons.__doc__.strip(), argv=argv))

    from localstack_ext.bootstrap import local_daemon

    if args["<subcommand>"] in [None, "start"]:
        print("Starting local daemons processes ...")
        local_daemon.start_in_background()
    elif args["<subcommand>"] in ["stop"]:
        print("Stopping local daemons processes ...")
        local_daemon.kill_servers()


def cmd_pod(argv, args):
    """
    Usage:
      localstack pod <subcommand> [ <name> ]

    Commands:
      push <name>      Push the state of a local cloud pod
      pull <name>      Pull the state to launch a cloud pod locally
    """
    args.update(docopt(cmd_pod.__doc__.strip(), argv=argv))
    from localstack_ext.bootstrap import (
        pods_client,  # leave import here to avoid dependency conflict
    )

    if args["<subcommand>"] == "push":
        print("Pushing local persistence state to backend ...")
        pods_client.push_state(args["<name>"], args)
    elif args["<subcommand>"] == "pull":
        print("Pulling persistence state from backend ...")
        pods_client.pull_state(args["<name>"], args)
    elif args["<subcommand>"] == "list":
        print("Fetching list of pods from server ...")
        result = pods_client.list_pods(args)
        # print('result', result)
        headers = {
            "pod_name": "Name",
            "backend": "Backend",
            "url": "URL",
            "size": "Size",
            "state": "State",
        }
        print_table(result, headers=headers, formats={"size": format_bytes})


def cmd_status(argv, args):
    from localstack.utils.cli import print_status

    print_status()
    print("\nExtended Details:")
    print("------------------")
    print("Pro Version:\t\t%s" % ext_constants.VERSION)
    # at this point we can assume that the API key has been validated...
    print("API Key:\t\tvalidated")


def register_commands():
    config.CLI_COMMANDS["daemons"] = {
        "description": "Manage local daemon processes",
        "parameters": [],
        "function": cmd_daemons,
    }
    config.CLI_COMMANDS["pod"] = {
        "description": "Manage state of local cloud pods",
        "parameters": [],
        "function": cmd_pod,
    }
    if "status" in config.CLI_COMMANDS:
        config.CLI_COMMANDS["status"]["function"] = cmd_status

    config.CLI_COMMANDS["login"] = {
        "description": "Log in with your account credentials",
        "function": cmd_login,
    }
    config.CLI_COMMANDS["logout"] = {
        "description": "Log out and delete any session tokens",
        "function": cmd_logout,
    }

    # TODO enable commands below?
    return

    config.CLI_COMMANDS["config"] = {
        "command": "  localstack config",
        "description": "Manage configuration values",
        "parameters": [],
        "function": cmd_config,
    }
    config.CLI_COMMANDS["ci"] = {
        "description": "Manage continuous integration repositories and settings",
        "function": cmd_ci,
    }


# -----------------
# HELPER FUNCTIONS
# -----------------


def mand_params(args, names=[]):
    return get_params(args, names, mandatory=True)


def opt_params(args, names=[]):
    return get_params(args, names, mandatory=False)


def get_params(args, names=[], mandatory=True):
    result = ()
    for name in names:
        value = None
        if name in args:
            value = args[name]
        if not value and mandatory:
            raise Exception("Please provide %s=... parameter" % name)
        result += (value,)
    if len(result) == 1:
        return result[0]
    return result


def print_table(data_list, headers=[], formats={}):
    from tabulate import tabulate

    if not isinstance(data_list, list):
        raise Exception("Non-list input cannot be rendered as table: %s" % data_list)
    formats = formats or {}
    data_list = [i if isinstance(i, dict) else {"value": i} for i in data_list]
    # prepare headers
    if isinstance(headers, dict):
        headers = [{"key": k, "name": v} for k, v in headers.items()]
    # build column metadata
    columns = headers
    if not columns:
        columns = []
        for row in data_list:
            for key in row.keys():
                if key not in columns:
                    columns.append(key)
    for idx in range(len(columns)):
        column = columns[idx]
        columns[idx] = column if isinstance(column, dict) else {"key": column, "name": column}

    # build rows
    def _format(row, key):
        value = row.get(key)
        formatter = formats.get(key)
        if formatter:
            return formatter(value)
        if not value:
            return value
        return truncate(str(value), max_length=30)

    rows = [[_format(row, h["key"]) for h in columns] for row in data_list]
    headers = [col["name"] for col in columns]
    result = tabulate(rows, headers=headers)
    print(result)



import logging
QthXf=Exception
QthXK=any
QthXr=str
QthXI=None
QthXO=False
QthXo=bool
QthXL=True
import os
from localstack import config as localstack_config
from localstack.constants import LOCALSTACK_INFRA_PROCESS,LOCALSTACK_WEB_PROCESS,TRUE_STRINGS
from localstack.utils import common
from localstack.utils.bootstrap import API_DEPENDENCIES,is_api_enabled
from localstack_ext import config as config_ext
from localstack_ext.bootstrap import cli,install,licensing,local_daemon
LOG=logging.getLogger(__name__)
EXTERNAL_PORT_APIS=("apigateway","apigatewayv2","athena","cloudfront","codecommit","ecs","ecr","elasticache","mediastore","rds","transfer","kafka","neptune","azure")
def register_localstack_plugins():
 _setup_logging()
 is_infra_process=os.environ.get(LOCALSTACK_INFRA_PROCESS)in TRUE_STRINGS
 is_api_key_configured=api_key_configured()
 if is_infra_process:
  install.install_libs()
  if is_api_key_configured:
   install.setup_ssl_cert()
 if os.environ.get(LOCALSTACK_WEB_PROCESS):
  return{}
 with licensing.prepare_environment():
  try:
   from localstack_ext.services import dns_server
   dns_server.setup_network_configuration()
  except QthXf:
   return
  if is_infra_process:
   load_plugin_files()
  try:
   if is_api_key_configured and not is_infra_process and is_api_enabled("ec2"):
    local_daemon.start_in_background()
  except QthXf as e:
   LOG.warning("Unable to start local daemon process: %s"%e)
  if is_api_key_configured:
   if os.environ.get("EDGE_PORT")and not localstack_config.EDGE_PORT_HTTP:
    LOG.warning(("!! Configuring EDGE_PORT={p} without setting EDGE_PORT_HTTP may lead "+"to issues; better leave the defaults, or set EDGE_PORT=443 and EDGE_PORT_HTTP={p}").format(p=localstack_config.EDGE_PORT))
   else:
    port=localstack_config.EDGE_PORT
    localstack_config.EDGE_PORT=443
    localstack_config.EDGE_PORT_HTTP=port
 API_DEPENDENCIES["apigateway"]=["apigatewayv2"]
 API_DEPENDENCIES["athena"]=["emr"]
 API_DEPENDENCIES["docdb"]=["rds"]
 API_DEPENDENCIES["ecs"]=["ecr"]
 API_DEPENDENCIES["elasticache"]=["ec2"]
 API_DEPENDENCIES["elb"]=["elbv2"]
 API_DEPENDENCIES["emr"]=["athena","s3"]
 API_DEPENDENCIES["glacier"]=["s3"]
 API_DEPENDENCIES["glue"]=["rds"]
 API_DEPENDENCIES["iot"]=["iotanalytics","iot-data","iotwireless"]
 API_DEPENDENCIES["kinesisanalytics"]=["kinesis","dynamodb"]
 API_DEPENDENCIES["neptune"]=["rds"]
 API_DEPENDENCIES["rds"]=["rds-data"]
 API_DEPENDENCIES["redshift"]=["redshift-data"]
 API_DEPENDENCIES["timestream"]=["timestream-write","timestream-query"]
 API_DEPENDENCIES["transfer"]=["s3"]
 docker_flags=[]
 if config_ext.use_custom_dns():
  if not common.is_port_open(dns_server.DNS_PORT,protocols="tcp"):
   docker_flags+=["-p {a}:{p}:{p}".format(a=config_ext.DNS_ADDRESS,p=dns_server.DNS_PORT)]
  if not common.is_port_open(dns_server.DNS_PORT,protocols="udp"):
   docker_flags+=["-p {a}:{p}:{p}/udp".format(a=config_ext.DNS_ADDRESS,p=dns_server.DNS_PORT)]
 if QthXK([is_api_enabled(api)for api in EXTERNAL_PORT_APIS]):
  docker_flags+=["-p {start}-{end}:{start}-{end}".format(start=config_ext.SERVICE_INSTANCES_PORTS_START,end=config_ext.SERVICE_INSTANCES_PORTS_END)]
 if is_api_enabled("eks"):
  kube_config=os.path.expanduser("~/.kube/config")
  if os.path.exists(kube_config):
   docker_flags+=["-v %s:/root/.kube/config"%kube_config]
 if is_api_enabled("azure"):
  docker_flags+=["-p {port}:{port}".format(port=5671)]
 if os.environ.get("AZURE"):
  docker_flags+=["-p {p}:{p}".format(p=config_ext.PORT_AZURE)]
 result={"docker":{"run_flags":" ".join(docker_flags)}}
 return result
def load_plugin_files():
 try:
  from localstack.services.plugins import Plugin,register_plugin
  from localstack_ext.bootstrap.dashboard import dashboard_extended
  from localstack_ext.services import edge
  from localstack_ext.services.amplify import amplify_starter
  from localstack_ext.services.apigateway import apigateway_extended
  from localstack_ext.services.applicationautoscaling import(applicationautoscaling_listener,applicationautoscaling_starter)
  from localstack_ext.services.appsync import appsync_starter
  from localstack_ext.services.athena import athena_starter
  from localstack_ext.services.autoscaling import autoscaling_starter
  from localstack_ext.services.awslambda import lambda_extended
  from localstack_ext.services.azure import azure_starter
  from localstack_ext.services.backup import backup_starter
  from localstack_ext.services.batch import batch_listener,batch_starter
  from localstack_ext.services.cloudformation import cloudformation_extended
  from localstack_ext.services.cloudfront import cloudfront_starter
  from localstack_ext.services.cloudtrail import cloudtrail_starter
  from localstack_ext.services.codecommit import codecommit_listener,codecommit_starter
  from localstack_ext.services.cognito import(cognito_identity_api,cognito_idp_api,cognito_starter)
  from localstack_ext.services.costexplorer import costexplorer_starter
  from localstack_ext.services.docdb import docdb_api
  from localstack_ext.services.dynamodb import dynamodb_extended
  from localstack_ext.services.ecr import ecr_listener,ecr_starter
  from localstack_ext.services.ecs import ecs_listener,ecs_starter
  from localstack_ext.services.efs import efs_api
  from localstack_ext.services.eks import eks_starter
  from localstack_ext.services.elasticache import elasticache_starter
  from localstack_ext.services.elasticbeanstalk import elasticbeanstalk_starter
  from localstack_ext.services.elb import elb_listener,elb_starter
  from localstack_ext.services.emr import emr_listener,emr_starter
  from localstack_ext.services.events import events_extended
  from localstack_ext.services.glacier import glacier_listener,glacier_starter
  from localstack_ext.services.glue import glue_listener,glue_starter
  from localstack_ext.services.iam import iam_extended
  from localstack_ext.services.iot import iot_listener,iot_starter
  from localstack_ext.services.kafka import kafka_starter
  from localstack_ext.services.kinesisanalytics import kinesis_analytics_api
  from localstack_ext.services.kms import kms_listener,kms_starter
  from localstack_ext.services.lakeformation import lakeformation_api
  from localstack_ext.services.mediastore import mediastore_starter
  from localstack_ext.services.neptune import neptune_api
  from localstack_ext.services.organizations import organizations_starter
  from localstack_ext.services.qldb import qldb_starter
  from localstack_ext.services.rds import rds_listener,rds_starter
  from localstack_ext.services.redshift import redshift_listener,redshift_starter
  from localstack_ext.services.route53 import route53_extended
  from localstack_ext.services.s3 import s3_extended
  from localstack_ext.services.sagemaker import sagemaker_starter
  from localstack_ext.services.secretsmanager import secretsmanager_extended
  from localstack_ext.services.serverlessrepo import serverlessrepo_starter
  from localstack_ext.services.servicediscovery import servicediscovery_starter
  from localstack_ext.services.ses import ses_extended
  from localstack_ext.services.sns import sns_extended
  from localstack_ext.services.sqs import sqs_extended
  from localstack_ext.services.stepfunctions import stepfunctions_extended
  from localstack_ext.services.sts import sts_extended
  from localstack_ext.services.timestream import timestream_starter
  from localstack_ext.services.transfer import transfer_starter
  from localstack_ext.services.xray import xray_listener,xray_starter
  from localstack_ext.utils import persistence as persistence_ext
  from localstack_ext.utils.aws import aws_utils
  register_plugin(Plugin("amplify",start=amplify_starter.start_amplify))
  register_plugin(Plugin("application-autoscaling",start=applicationautoscaling_starter.start_applicationautoscaling,listener=applicationautoscaling_listener.UPDATE_APPLICATION_AUTOSCALING))
  register_plugin(Plugin("appsync",start=appsync_starter.start_appsync))
  register_plugin(Plugin("athena",start=athena_starter.start_athena))
  register_plugin(Plugin("autoscaling",start=autoscaling_starter.start_autoscaling))
  register_plugin(Plugin("azure",start=azure_starter.start_azure))
  register_plugin(Plugin("backup",start=backup_starter.start_backup))
  register_plugin(Plugin("batch",start=batch_starter.start_batch,listener=batch_listener.UPDATE_BATCH))
  register_plugin(Plugin("cloudfront",start=cloudfront_starter.start_cloudfront))
  register_plugin(Plugin("cloudtrail",start=cloudtrail_starter.start_cloudtrail))
  register_plugin(Plugin("codecommit",start=codecommit_starter.start_codecommit,listener=codecommit_listener.UPDATE_CODECOMMIT))
  register_plugin(Plugin("cognito-identity",start=cognito_starter.start_cognito_identity,listener=cognito_identity_api.UPDATE_COGNITO_IDENTITY))
  register_plugin(Plugin("cognito-idp",start=cognito_starter.start_cognito_idp,listener=cognito_idp_api.UPDATE_COGNITO))
  register_plugin(Plugin("ce",start=costexplorer_starter.start_costexplorer))
  register_plugin(Plugin("docdb",start=docdb_api.start_docdb))
  register_plugin(Plugin("efs",start=efs_api.start_efs))
  register_plugin(Plugin("elasticache",start=elasticache_starter.start_elasticache))
  register_plugin(Plugin("elasticbeanstalk",start=elasticbeanstalk_starter.start_elasticbeanstalk))
  register_plugin(Plugin("elb",start=elb_starter.start_elb,listener=elb_listener.UPDATE_ELB))
  register_plugin(Plugin("elbv2",start=elb_starter.start_elbv2,listener=elb_listener.UPDATE_ELB))
  register_plugin(Plugin("ecr",start=ecr_starter.start_ecr,listener=ecr_listener.UPDATE_ECR))
  register_plugin(Plugin("ecs",start=ecs_starter.start_ecs,listener=ecs_listener.UPDATE_ECS))
  register_plugin(Plugin("eks",start=eks_starter.start_eks))
  register_plugin(Plugin("emr",start=emr_starter.start_emr,listener=emr_listener.UPDATE_EMR))
  register_plugin(Plugin("glacier",start=glacier_starter.start_glacier,listener=glacier_listener.UPDATE_GLACIER))
  register_plugin(Plugin("glue",start=glue_starter.start_glue,listener=glue_listener.UPDATE_GLUE))
  register_plugin(Plugin("iot",start=iot_starter.start_iot,listener=iot_listener.UPDATE_IOT))
  register_plugin(Plugin("kafka",start=kafka_starter.start_kafka))
  register_plugin(Plugin("kinesisanalytics",start=kinesis_analytics_api.start_kinesis_analytics))
  register_plugin(Plugin("kms",start=kms_starter.start_kms,listener=kms_listener.UPDATE_KMS))
  register_plugin(Plugin("lakeformation",start=lakeformation_api.start_lakeformation))
  register_plugin(Plugin("mediastore",start=mediastore_starter.start_mediastore))
  register_plugin(Plugin("neptune",start=neptune_api.start_neptune))
  register_plugin(Plugin("organizations",start=organizations_starter.start_organizations))
  register_plugin(Plugin("qldb",start=qldb_starter.start_qldb))
  register_plugin(Plugin("rds",start=rds_starter.start_rds,listener=rds_listener.UPDATE_RDS))
  register_plugin(Plugin("redshift",start=redshift_starter.start_redshift,listener=redshift_listener.UPDATE_REDSHIFT,priority=10))
  register_plugin(Plugin("sagemaker",start=sagemaker_starter.start_sagemaker))
  register_plugin(Plugin("serverlessrepo",start=serverlessrepo_starter.start_serverlessrepo))
  register_plugin(Plugin("servicediscovery",start=servicediscovery_starter.start_servicediscovery))
  register_plugin(Plugin("timestream",start=timestream_starter.start_timestream))
  register_plugin(Plugin("transfer",start=transfer_starter.start_transfer))
  register_plugin(Plugin("xray",start=xray_starter.start_xray,listener=xray_listener.UPDATE_XRAY))
  persistence_ext.enable_extended_persistence()
  lambda_extended.patch_lambda()
  sns_extended.patch_sns()
  sqs_extended.patch_sqs()
  apigateway_extended.patch_apigateway()
  cloudformation_extended.patch_cloudformation()
  stepfunctions_extended.patch_stepfunctions()
  s3_extended.patch_s3()
  iam_extended.patch_iam()
  dynamodb_extended.patch_dynamodb()
  events_extended.patch_events()
  secretsmanager_extended.patch_secretsmanager()
  sts_extended.patch_sts()
  ses_extended.patch_ses()
  route53_extended.patch_route53()
  dashboard_extended.patch_dashboard()
  aws_utils.patch_aws_stack()
  edge.patch_start_edge()
  patch_start_infra()
 except QthXf as e:
  if "No module named" not in QthXr(e):
   print("ERROR: %s"%e)
def patch_start_infra():
 from localstack.services import infra
 try:
  from localstack_ext.utils.aws.metadata_service import start_metadata_service
 except QthXf:
  start_metadata_service=QthXI
 def do_start_infra(asynchronous,apis,is_in_docker,*args,**kwargs):
  if common.in_docker():
   try:
    start_metadata_service and start_metadata_service()
   except QthXf:
    pass
  enforce_before=config_ext.ENFORCE_IAM
  try:
   config_ext.ENFORCE_IAM=QthXO
   return do_start_infra_orig(asynchronous,apis,is_in_docker,*args,**kwargs)
  finally:
   config_ext.ENFORCE_IAM=enforce_before
 do_start_infra_orig=infra.do_start_infra
 infra.do_start_infra=do_start_infra
def _setup_logging():
 log_level=logging.DEBUG if localstack_config.DEBUG else logging.INFO
 logging.getLogger("localstack_ext").setLevel(log_level)
 logging.getLogger("botocore").setLevel(logging.INFO)
 logging.getLogger("kubernetes").setLevel(logging.INFO)
 logging.getLogger("pyftpdlib").setLevel(logging.INFO)
 logging.getLogger("pyhive").setLevel(logging.INFO)
 logging.getLogger("websockets").setLevel(logging.INFO)
 logging.getLogger("asyncio").setLevel(logging.INFO)
 logging.getLogger("hpack").setLevel(logging.INFO)
 logging.getLogger("jnius.reflect").setLevel(logging.INFO)
 logging.getLogger("dulwich").setLevel(logging.ERROR)
 logging.getLogger("postgresql_proxy").setLevel(logging.WARNING)
 logging.getLogger("intercept").setLevel(logging.WARNING)
 logging.getLogger("root").setLevel(logging.WARNING)
 logging.getLogger("").setLevel(logging.WARNING)
def api_key_configured():
 return QthXo(os.environ.get("LOCALSTACK_API_KEY"))
def register_localstack_commands():
 if api_key_configured():
  cli.register_commands()
 return QthXL
# Created by pyminifier (https://github.com/liftoff/pyminifier)



#!/usr/bin/env python

import os
import re
import sys

from setuptools import find_packages, setup

# marker for basic and basic libs in requirements.txt
BASIC_LIB_MARKER = '#basic-lib'

# libs that require Python 3.x
PYTHON3_LIBS = ['hbmqtt', 'graphql-core', 'websockets']

install_requires = []
dependency_links = []
extra_requires = []

# root code folder
THIS_FOLDER = os.path.realpath(os.path.dirname(__file__))


# determine version
with open(os.path.join(THIS_FOLDER, 'localstack_ext', 'constants.py')) as f:
    configs = f.read()
version = re.search(r'^\s*VERSION\s*=\s*[\'"](.+)[\'"]\s*$', configs, re.MULTILINE).group(1)

# read requirements
with open(os.path.join(THIS_FOLDER, 'requirements.txt')) as f:
    requirements = f.read()

for line in re.split('\n', requirements):
    if line and line[0] == '#' and '#egg=' in line:
        line = re.search(r'#\s*(.*)', line).group(1)
    if line and line[0] != '#':
        if '://' not in line:
            lib_stripped = line.split(' #')[0].strip()
            if BASIC_LIB_MARKER in line:
                install_requires.append(lib_stripped)
            else:
                extra_requires.append(lib_stripped)

if sys.version_info[:2] < (3, 4):
    # remove libs that require Python 3+
    def requires_p3(lib):
        return any([p3l in lib for p3l in PYTHON3_LIBS])
    install_requires = [d for d in install_requires if not requires_p3(d)]
    extra_requires = [d for d in extra_requires if not requires_p3(d)]


package_data = {
    '': ['*.md'],
    'localstack_ext': [
        'utils/*.py.enc',
        'utils/aws/*.py.enc',
        'services/*.py.enc',
        'services/amplify/*.py.enc',
        'services/apigateway/*.py.enc',
        'services/applicationautoscaling/*.py.enc',
        'services/appsync/*.py.enc',
        'services/athena/*.py.enc',
        'services/autoscaling/*.py.enc',
        'services/awslambda/*.py.enc',
        'services/azure/*.py.enc',
        'services/backup/*.py.enc',
        'services/batch/*.py.enc',
        'services/cloudformation/*.py.enc',
        'services/cloudfront/*.py.enc',
        'services/cloudtrail/*.py.enc',
        'services/codecommit/*.py.enc',
        'services/cognito/*.json',
        'services/cognito/*.py.enc',
        'services/costexplorer/*.py.enc',
        'services/docdb/*.py.enc',
        'services/dynamodb/*.py.enc',
        'services/ec2/*.py.enc',
        'services/ecr/*.py.enc',
        'services/ecs/*.py.enc',
        'services/efs/*.py.enc',
        'services/eks/*.py.enc',
        'services/elb/*.py.enc',
        'services/emr/*.py.enc',
        'services/elasticache/*.py.enc',
        'services/elasticbeanstalk/*.py.enc',
        'services/events/*.py.enc',
        'services/glacier/*.py.enc',
        'services/glue/*.py.enc',
        'services/iam/*.json',
        'services/iam/*.py.enc',
        'services/iot/*.py.enc',
        'services/kafka/*.py.enc',
        'services/kinesisanalytics/*.py.enc',
        'services/kms/*.py.enc',
        'services/lakeformation/*.py.enc',
        'services/mediastore/*.py.enc',
        'services/neptune/*.py.enc',
        'services/organizations/*.py.enc',
        'services/qldb/*.py.enc',
        'services/rds/*.py.enc',
        'services/redshift/*.py.enc',
        'services/route53/*.py.enc',
        'services/s3/*.py.enc',
        'services/sagemaker/*.py.enc',
        'services/secretsmanager/*.py.enc',
        'services/serverlessrepo/*.yml',
        'services/serverlessrepo/*.py.enc',
        'services/servicediscovery/*.py.enc',
        'services/ses/*.py.enc',
        'services/sns/*.py.enc',
        'services/sqs/*.py.enc',
        'services/stepfunctions/*.py.enc',
        'services/sts/*.py.enc',
        'services/timestream/*.py.enc',
        'services/transfer/*.py.enc',
        'services/xray/*.py.enc'
    ]}


if __name__ == '__main__':

    setup(
        name='localstack-ext',
        version=version,
        description='Extensions for LocalStack',
        author='Waldemar Hummer',
        author_email='waldemar.hummer@gmail.com',
        url='https://github.com/localstack/localstack',
        packages=find_packages(exclude=('tests', 'tests.*')),
        package_data=package_data,
        install_requires=install_requires,
        dependency_links=dependency_links,
        extras_require={
            'full': extra_requires
        },
        test_suite='tests',
        zip_safe=False,
        classifiers=[
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'License :: OSI Approved :: Apache Software License',
            'Topic :: Software Development :: Testing',
        ]
    )



#!/usr/bin/env python
iIsYG=dict
iIsYK=super
iIsYx=hasattr
iIsYA=Exception
iIsYB=object
iIsYF=dir
iIsYg=None
iIsYE=False
iIsYL=True
iIsYu=len
iIsYr=set
import errno
import socket
from select import select
from socket import(AF_INET,AF_INET6,IPPROTO_IPV6,IPPROTO_TCP,IPV6_V6ONLY,SO_REUSEADDR,SOCK_STREAM,SOL_SOCKET,TCP_NODELAY)
class AttrDict(iIsYG):
 def __init__(self,*args,**kwargs):
  iIsYK(AttrDict,self).__init__(*args,**kwargs)
  self.__dict__=self
def tcp_listen(six,addr,port,blk):
 s=socket.socket(AF_INET6 if six else AF_INET,SOCK_STREAM)
 if six and iIsYx(socket,"IPV6_V6ONLY"):
  s.setsockopt(IPPROTO_IPV6,IPV6_V6ONLY,0)
 s.setsockopt(SOL_SOCKET,SO_REUSEADDR,1)
 s.setsockopt(IPPROTO_TCP,TCP_NODELAY,1)
 s.bind((addr,port))
 s.listen(5)
 s.setblocking(blk)
 return s
def tcp_connect(six,addr,port,blk):
 s=socket.socket(AF_INET6 if six else AF_INET,SOCK_STREAM)
 s.setsockopt(IPPROTO_TCP,TCP_NODELAY,1)
 s.connect((addr,port))
 s.setblocking(blk)
 return s
def safe_close(x):
 try:
  x.close()
 except iIsYA:
  pass
class Server(iIsYB):
 def __init__(self,opt,q):
  self.opt=opt
  self.sock=tcp_listen(opt.ip6,opt.bind_addr,opt.bind_port,0)
  self.q=q
 def pre_wait(self,rr,r,w,e):
  r.append(self.sock)
 def post_wait(self,r,w,e):
  if self.sock in r:
   cl,addr=self.sock.accept()
   cl.setblocking(0)
   self.q.append(Proxy(self.opt,cl,addr))
class Half(iIsYB):
 def __init__(self,opt,sock,addr,iIsYF):
  self.opt=opt
  self.sock=sock
  self.addr=addr
  self.iIsYF=iIsYF
  self.name="peer" if self.iIsYF else "client"
  self.queue=[]
  self.dest=iIsYg
  self.err=iIsYg
  self.ready=iIsYE
 def pre_wait(self,rr,r,w,e):
  if self.ready:
   rr.append(self.sock)
  r.append(self.sock)
  if self.queue:
   w.append(self.sock)
 def post_wait(self,r,w,e):
  if not self.err and self.sock in w and self.queue:
   self.write_some()
  if not self.err and self.sock in r:
   self.ready=iIsYL
   self.copy()
  return self.err
 def error(self,msg,e):
  self.err="Error on %s: %s: %s"%(self.name,msg,e)
  return self.err
 def write_some(self):
  try:
   n=self.sock.send(self.queue[0])
  except iIsYA as e:
   return self.error("send error",e)
  if n!=iIsYu(self.queue[0]):
   self.queue[0]=self.queue[0][n:]
  else:
   del self.queue[0]
 def copy(self):
  try:
   buf=self.sock.recv(4096)
  except socket.error as e:
   if e.errno==errno.EWOULDBLOCK:
    self.ready=iIsYE
    return
   return self.error("recv socket error",e)
  except iIsYA as e:
   return self.error("recv error",e)
  if iIsYu(buf)==0:
   return self.error("eof",0)
  self.dest.queue.append(buf)
 def close(self):
  safe_close(self.sock)
class Proxy(iIsYB):
 def __init__(self,opt,sock,addr):
  self.opt=opt
  self.cl=Half(opt,sock,addr,"i")
  peer=tcp_connect(opt.ip6,opt.addr,opt.port,0)
  self.peer=Half(opt,peer,addr,"o")
  self.cl.dest=self.peer
  self.peer.dest=self.cl
  self.err=iIsYg
 def pre_wait(self,rr,r,w,e):
  self.cl.pre_wait(rr,r,w,e)
  self.peer.pre_wait(rr,r,w,e)
 def post_wait(self,r,w,e):
  if not self.err:
   self.err=self.cl.post_wait(r,w,e)
  if not self.err:
   self.err=self.peer.post_wait(r,w,e)
  if self.err:
   self.cl.close()
   self.peer.close()
  return self.err
def server_loop(opt):
 qs=[]
 opt["ip6"]=opt.get("ip6",iIsYE)
 opt["bind_addr"]=opt.get("bind_addr","0.0.0.0")
 opt["addr"]=opt.get("addr","127.0.0.1")
 opt["bind_port"]=opt.get("bind_port",0)
 opt=AttrDict(opt)
 qs.append(Server(opt,qs))
 while qs:
  rr,r,w,e=[],[],[],[]
  for q in qs:
   q.pre_wait(rr,r,w,e)
  timeo=10.0 if not rr else 0.0
  r,w,e=select(r,w,e,timeo)
  r=iIsYr(r).union(rr)
  for q in qs:
   if q.post_wait(r,w,e):
    qs.remove(q)
def main():
 opt={"ip6":iIsYE,"bind_port":2223,"addr":"192.168.43.191","port":22}
 opt={"ip6":iIsYE,"bind_port":48429,"addr":"128.0.100.3","port":22}
 opt={"ip6":iIsYE,"port":54237,"bind_addr":"128.0.100.2","bind_port":22}
 opt={"ip6":iIsYE,"port":54237,"bind_addr":"192.168.100.3","bind_port":22}
 server_loop(opt)
if __name__=="__main__":
 main()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import base64
uvcng=True
uvcnO=Exception
uvcno=False
uvcnX=str
uvcnC=len
uvcnp=int
uvcnD=range
uvcnt=object
import glob
import json
import logging
import os
import re
import sys
import traceback
import pyaes
from localstack import config as localstack_config
from localstack.constants import ENV_PRO_ACTIVATED
from localstack.utils.common import load_file,md5,now_utc,parallelize
from localstack.utils.common import safe_requests as requests
from localstack.utils.common import save_file,str_insert,str_remove,to_bytes,to_str
from localstack_ext import config
from localstack_ext.config import PROTECTED_FOLDERS,ROOT_FOLDER
from localstack_ext.constants import VERSION
ENV_PREPARED={}
MAX_KEY_CACHE_DURATION_SECS=60*60*24
LOG=logging.getLogger(__name__)
ENV_LOCALSTACK_API_KEY="LOCALSTACK_API_KEY"
def read_api_key(raise_if_missing=uvcng):
 key=(os.environ.get(ENV_LOCALSTACK_API_KEY)or "").strip()
 if not key and raise_if_missing:
  raise uvcnO("Unable to retrieve API key. Please configure $%s in your environment"%ENV_LOCALSTACK_API_KEY)
 return key
def fetch_key():
 api_key=read_api_key()
 if api_key=="test":
  return "test"
 data={"api_key":api_key,"version":VERSION}
 try:
  logging.getLogger("py.warnings").setLevel(logging.ERROR)
  result=requests.post("%s/activate"%config.API_URL,json.dumps(data),verify=uvcno)
  if result.status_code>=400:
   content=result.content
   content_type=result.headers.get("Content-Type")
   raise uvcnO('Received error activating key (code %s): ctype "%s" - %s'%(result.status_code,content_type,content))
  key_base64=json.loads(to_str(result.content))["key"]
  cache_key_locally(api_key,key_base64)
 except uvcnO as e:
  if log_license_issues():
   api_key=uvcnX(api_key_configured()or "")
   api_key_prefix=api_key[:3]
   LOG.warning('Error activating API key "%s..."(%s): %s %s'%(api_key_prefix,uvcnC(api_key),e,traceback.format_exc()))
  key_base64=load_cached_key(api_key)
 finally:
  logging.getLogger("py.warnings").setLevel(logging.WARNING)
 decoded_key=to_str(base64.b64decode(key_base64))
 return decoded_key
def cache_key_locally(api_key,key_b64):
 configs=localstack_config.load_config_file()
 timestamp=uvcnX(uvcnp(now_utc()))
 key_raw=to_str(base64.b64decode(key_b64))
 for i in uvcnD(uvcnC(timestamp)):
  key_raw=str_insert(key_raw,i*2,timestamp[i])
 key_b64=to_str(base64.b64encode(to_bytes(key_raw)))
 configs["cached_key"]={"timestamp":uvcnp(timestamp),"key_hash":md5(api_key),"key":key_b64}
 save_file(localstack_config.CONFIG_FILE_PATH,json.dumps(configs))
 return configs
def load_cached_key(api_key):
 configs=localstack_config.load_config_file()
 cached_key=configs["cached_key"]
 if cached_key.get("key_hash")!=md5(api_key):
  raise uvcnO("Cached key was created for a different API key")
 now=now_utc()
 if(now-cached_key["timestamp"])>MAX_KEY_CACHE_DURATION_SECS:
  raise uvcnO("Cached key expired")
 timestamp=uvcnX(cached_key["timestamp"])
 key_raw=to_str(base64.b64decode(cached_key["key"]))
 for i in uvcnD(uvcnC(timestamp)):
  assert key_raw[i]==timestamp[i]
  key_raw=str_remove(key_raw,i)
 key_b64=to_str(base64.b64encode(to_bytes(key_raw)))
 return key_b64
def generate_aes_cipher(key):
 key=to_bytes(key)
 return pyaes.AESModeOfOperationCBC(key,iv="\0"*16)
def decrypt_file(source,target,key):
 cipher=generate_aes_cipher(key)
 raw=load_file(source,mode="rb")
 decrypter=pyaes.Decrypter(cipher)
 decrypted=decrypter.feed(raw)
 decrypted+=decrypter.feed()
 decrypted=decrypted.partition(b"\0")[0]
 decrypted=to_str(decrypted)
 save_file(target,content=decrypted)
def decrypt_files(key):
 files=[]
 for folder in PROTECTED_FOLDERS:
  for subpath in("*.py.enc","**/*.py.enc"):
   for f in glob.glob("%s/localstack_ext/%s/%s"%(ROOT_FOLDER,folder,subpath)):
    files.append(f)
 def _decrypt(f):
  target=f[:-4]
  if not os.path.exists(target):
   decrypt_file(f,target,key)
 parallelize(_decrypt,files)
def cleanup_environment():
 excepted_files=r".*/services/((edge)|(dns_server)|(__init__))\.py"
 for folder in PROTECTED_FOLDERS:
  for subpath in("*.py.enc","**/*.py.enc"):
   for f in glob.glob("%s/localstack_ext/%s/%s"%(ROOT_FOLDER,folder,subpath)):
    target=f[:-4]
    if not re.match(excepted_files,target):
     for delete_file in(target,"%sc"%target):
      if os.path.exists(delete_file):
       os.remove(delete_file)
def prepare_environment():
 class OnClose(uvcnt):
  def __exit__(self,*args,**kwargs):
   if not ENV_PREPARED.get("finalized"):
    cleanup_environment()
   ENV_PREPARED["finalized"]=uvcng
  def __enter__(self,*args,**kwargs):
   pass
 if not ENV_PREPARED.get("finalized"):
  try:
   key=fetch_key()
   if not key:
    raise uvcnO("Unable to fetch and validate API key from environment")
   if key!="test":
    decrypt_files(key)
    LOG.info("Successfully activated API key")
   else:
    LOG.info("Using test API key")
   os.environ[ENV_PRO_ACTIVATED]="1"
  except uvcnO as e:
   if log_license_issues():
    LOG.warning("Unable to activate API key: %s %s"%(e,traceback.format_exc()))
   if config.REQUIRE_PRO:
    LOG.error("Unable to activate API key, but $REQUIRE_PRO is configured - quitting.")
    sys.exit(1)
 return OnClose()
def log_license_issues():
 return api_key_configured()and localstack_config.is_env_true("LOG_LICENSE_ISSUES")
def api_key_configured():
 return read_api_key(raise_if_missing=uvcno)
def get_auth_headers():
 configs=localstack_config.load_config_file()
 login_info=configs.get("login")
 if not login_info:
  raise uvcnO("Please log in first")
 auth_token=login_info["token"]
 if not auth_token.startswith("%s "%login_info["provider"]):
  auth_token="%s %s"%(login_info["provider"],auth_token)
 headers={"authorization":auth_token}
 return headers
# Created by pyminifier (https://github.com/liftoff/pyminifier)
